{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE ImportQualifiedPost #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE RecursiveDo #-}

-- see https://gh.sudoer.ch/danieljharvey/mimsa/blob/trunk/llvm-calc/src/Calc/Compile/ToLLVM.hs
-- see https://blog.josephmorag.com/posts/mcc3/

module Main.LLVMGen (llvmGen) where

import Control.Monad.State
import Control.Monad.Trans.Class (lift)
import Data.ByteString (ByteString)
import Data.Map qualified as M
import Data.String.Conversions
import Data.Text
import LLVM (moduleLLVMAssembly, withModuleFromAST)
import LLVM.AST hiding (function)
import LLVM.AST.IntegerPredicate
import LLVM.AST.Type
import LLVM.AST.Type qualified as AST
import LLVM.Context
import LLVM.IRBuilder.Constant
import LLVM.IRBuilder.Instruction
import LLVM.IRBuilder.Module
import LLVM.IRBuilder.Monad
import Main.Types qualified as T

data Env = Env
  { operands :: M.Map Text Operand,
    strings :: M.Map Text Operand
  }
  deriving (Eq, Show)

registerOperand :: (MonadState Env m) => Text -> Operand -> m ()
registerOperand name op =
  modify $ \env -> env {operands = M.insert name op (operands env)}

registerString :: (MonadState Env m) => Text -> Operand -> m ()
registerString name op =
  modify $ \env -> env {strings = M.insert name op (operands env)}

getOperand :: (MonadState Env m, MonadModuleBuilder m, MonadIRBuilder m) => Text -> m Operand
getOperand name = do
  env <- get
  case M.lookup name (operands env) of
    Just op -> return op
    Nothing -> error $ "Unknown operand: " ++ show name

getString :: (MonadState Env m, MonadModuleBuilder m, MonadIRBuilder m) => Text -> m Operand
getString str = do
  env <- get
  case M.lookup str (strings env) of
    Just s -> return s
    Nothing -> do
      s <- globalStringPtr (unpack str) (mkName "str")
      let operand = ConstantOperand s
      modify $ \env -> env {strings = M.insert str operand (strings env)}
      return operand

getLLVM :: [T.Statement] -> Module
getLLVM statement =
  flip evalState (Env {operands = M.empty, strings = M.empty}) $
    buildModuleT "program" $ mdo
      -- TODO: better module name
      printf <- externVarArgs "printf" [ptr] i32
      lift $ registerOperand "printf" printf
      function "main" [] i32 $ \_ -> mdo
        printNumStr <- globalStringPtr "%d\n" (mkName "str")
        lift $ registerString "%d\n" $ ConstantOperand printNumStr
        _ <- forM_ statement statementToLLVM
        ret $ int32 0

--
-- ourExpression <- exprToLLVM expr
-- _ <- call (FunctionType i32 [ptr] True) printf [(ConstantOperand numFormatStr, []), (ourExpression, [])]
-- ret $ int32 0

statementToLLVM ::
  ( MonadIRBuilder m,
    MonadModuleBuilder m,
    MonadState Env m
  ) =>
  T.Statement ->
  m Operand
statementToLLVM (T.PrintInt e) = mdo
  val <- intExprToLLVM e
  printf <- getOperand "printf"
  formatStr <- getString "%d\n"
  _ <- call (FunctionType i32 [ptr] True) printf [(formatStr, []), (val, [])]
  pure val
statementToLLVM (T.PrintBool e) = mdo
  val <- boolExprToLLVM e
  val32 <- zext val i32
  printf <- getOperand "printf"
  formatStr <- getString "%d\n"
  _ <- call (FunctionType i32 [ptr] True) printf [(formatStr, []), (val32, [])]
  pure val

intExprToLLVM ::
  ( MonadIRBuilder m,
    MonadModuleBuilder m,
    MonadState Env m
  ) =>
  T.Int ->
  m Operand
intExprToLLVM (T.Int prim) = pure $ int32 $ fromIntegral prim
intExprToLLVM (T.IntArith T.Add a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  add lhs rhs
intExprToLLVM (T.IntArith T.Sub a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  sub lhs rhs
intExprToLLVM (T.IntArith T.Mul a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  mul lhs rhs
intExprToLLVM (T.IntArith T.Div a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  sdiv lhs rhs

boolExprToLLVM ::
  ( MonadIRBuilder m,
    MonadModuleBuilder m,
    MonadState Env m
  ) =>
  T.Bool ->
  m Operand
boolExprToLLVM (T.Bool prim) =
  if prim then pure $ bit 1 else pure $ bit 0
boolExprToLLVM (T.IntOrdCmp T.GT a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  icmp SGT lhs rhs
boolExprToLLVM (T.IntOrdCmp T.GTE a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  icmp SGE lhs rhs
boolExprToLLVM (T.IntOrdCmp T.LT a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  icmp SLT lhs rhs
boolExprToLLVM (T.IntOrdCmp T.LTE a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  icmp SLE lhs rhs
boolExprToLLVM (T.IntEq T.EQ a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  icmp LLVM.AST.IntegerPredicate.EQ lhs rhs
boolExprToLLVM (T.IntEq T.NE a b) = mdo
  lhs <- intExprToLLVM a
  rhs <- intExprToLLVM b
  icmp LLVM.AST.IntegerPredicate.NE lhs rhs

llvmGen :: [T.Statement] -> IO ByteString
llvmGen expr = do
  let l = getLLVM expr
  withContext $ \c -> withModuleFromAST c l moduleLLVMAssembly